import functools
import typing as t

from mb.commands import Command
from mb.events import Event
from mb.messages import Message
from mb.unit_of_work.base import UnitOfWork

UOW = t.TypeVar("UOW", bound=UnitOfWork)
M = t.TypeVar("M", bound=Message)
C = t.TypeVar("C", bound=Command)
E = t.TypeVar("E", bound=Event)
R = t.TypeVar("R")


class TrackedHandler(t.Protocol[M, R]):
    calls: list[M]
    __call__: t.Callable[["TrackedHandler", M, UOW], R]


def tracked_handler(func: t.Callable[[M, UOW], R]) -> TrackedHandler[M, R]:
    """
    Tracks the calls to a handler in the :attr:`calls`.
    """

    @functools.wraps(func)
    def wrapper(m: M, uow: UOW) -> R:
        wrapper.calls.append(m)  # type:ignore
        return func(m, uow)

    wrapper.calls = []  # type:ignore
    return wrapper  # type:ignore



# --------------------------------------
# Base exceptions
# --------------------------------------


class CQBaseError(BaseException):
    """
    Uncatchable error
    """


class CQProgrammingError(CQBaseError, RuntimeError):
    """
    Uncatchable error. Code violates some invariant
    """


class CQError(Exception):
    """
    Library root exception
    """


class ConfigError(CQError):
    """
    General configuration error
    """


# --------------------------------------
# Message definitions
# --------------------------------------


class InvalidMessageName(ConfigError, TypeError):
    """
    The Message.NAME must be an string
    """


class DuplicatedMessageName(ConfigError, TypeError):
    """
    Two messages have the same NAME
    """


# --------------------------------------
# Handler configuration
# --------------------------------------


class DuplicatedCommandHandler(ConfigError):
    """
    A Command can have only one handler
    """


class MissingCommandHandler(ConfigError, RuntimeError):
    """
    No handler found for a Command
    """


class InvalidMessageType(ConfigError, TypeError):
    """
    The Bus only handles Messages
    """


# --------------------------------------
# Handling messages
# --------------------------------------


class InvalidMessage(CQError, TypeError):
    """
    The Bus only handles Messages
    """


# --------------------------------------
# Databases
# --------------------------------------


class InvalidTransactionState(CQError):
    """
    Likely an internal error
    """


# --------------------------------------
# Unit of Work
# --------------------------------------


class UowContextRequired(CQError):
    """
    Code is using the UnitOfWork outside the context or
    the begin/commit/rollback calls are unpaired.

        >>> uow = UnitOfWork(bus)
        >>> with uow:
        ...     uow.emit_event(event)  # Good
        ... uow.emit_event(event)  # Bad
    """



import inspect
import logging
import typing as t
from collections import defaultdict

from mb.commands import Command
from mb.events import Event
from mb.exceptions import DuplicatedCommandHandler
from mb.exceptions import InvalidMessage
from mb.exceptions import InvalidMessageType
from mb.exceptions import MissingCommandHandler
from mb.unit_of_work import UnitOfWork

logger = logging.getLogger()

# --------------------------------------
# Typings
# --------------------------------------

#
# TODO: Force callbacks' first argument type (command or event) but allow
# arbitrary keyword arguments to implement Dependency Injection features.
# Not all handlers need the current UnitOfWork and others might need
# other global services to work (e.g. EmailsService) that could be
# injected.
#
# Currently, Python 3.9 does not offer the abilities do to so. Will be
# the new Python 3.10 features enough? Concatenate, Param, ...
#

UOW = t.TypeVar("UOW", bound=UnitOfWork)
C = t.TypeVar("C", bound=Command)
E = t.TypeVar("E", bound=Event)


class CommandHandlers(t.Protocol):  # pragma: no cover
    def __setitem__(self, key: type[C], item: t.Callable[[C, UOW], t.Any]):
        ...

    def __getitem__(self, item: type[C]) -> t.Callable[[C, UOW], t.Any]:
        ...


class EventHandlers(t.Protocol):  # pragma: no cover
    def __getitem__(self, item: type[E]) -> list[t.Callable[[E, UOW], t.Any]]:
        ...


# --------------------------------------
# Bus
# --------------------------------------

#
# NOTE: subscrbing to Event means subscrbing to all events. In the same way,
# subscribing to ParentEvent means subscribing to any ChildEvent(ParentEvent).
# I dunno if this makes sense for everyone, I'm just exploring options and
# trying in a real project to get a real experience.
# Btw, I'll better check/revise how this works in Kafka or Rabbit and get some
# clever ideas from the real pros.
#


class MessageBus:
    """
    An in-memory message bus.
    """

    event_handlers: EventHandlers
    command_handlers: CommandHandlers

    def __init__(self):
        self.event_handlers = defaultdict(list)
        self.command_handlers = {}

    def subscribe_event(
        self,
        cls: type[E],
        handler: t.Callable[[E, UOW], t.Any],
    ) -> None:
        """
        Subscribe to an event type. An event may have multiple handlers

        :raises InvalidMessageType:
        """
        if not issubclass(cls, Event):
            raise InvalidMessageType(f"This is not an event class: '{cls}'")
        self.event_handlers[cls].append(handler)

    def subscribe_command(
        self,
        cls: type[C],
        handler: t.Callable[[C, UOW], t.Any],
    ) -> None:
        """
        Set the command handler for this command. Only one handler allowed.

        :raises InvalidMessageType:
        :raises ConfigError: if there is already a handler
        """
        if not issubclass(cls, Command):
            raise InvalidMessageType(f"This is not a command class: '{cls}'")
        try:
            current_handler = self.command_handlers[cls]
        except KeyError:
            self.command_handlers[cls] = handler
        else:
            raise DuplicatedCommandHandler(
                f"Duplicated handler for command '{cls}'. "
                f"The handler '{handler}' overrides the current '{current_handler}'"
            )

    def _handle_command(self, command: C, uow: UnitOfWork) -> t.Any:
        """
        Triggers the command handler. Handler exceptions are propagated

        :raises InvalidMessage: if this isn't a Command
        :raises MissingCommandHandler: if there's no handler configured for the command
        """
        if not isinstance(command, Command):
            raise InvalidMessage(f"This is not an command: '{command}'")

        try:
            handler = self.command_handlers[type(command)]
        except KeyError:
            raise MissingCommandHandler(f"Missing handler for command: '{command}'")

        logger.debug(f"Handling command '{command}': calling '{handler}'")
        return handler(command, uow)

    def _handle_event(self, event: E, uow: UnitOfWork) -> None:
        """
        Triggers the event handlers. Handler exceptions are captured and error-logged.
        """
        handlers: list[t.Callable[[Event, UnitOfWork], None]] = []
        for event_cls in inspect.getmro(type(event)):
            handlers.extend(self.event_handlers[event_cls])
        logger.debug(f"Handling event '{event}': {len(handlers)} handlers")

        for handler in handlers:
            logger.debug(f"Handling event '{event}': calling '{handler}'")
            try:
                handler(event, uow)
            except Exception:
                logger.exception(f"Exception handling event '{event}'")

    def clone(self) -> "MessageBus":
        """
        Returns a copy of itself with the same (detached) configuration
        """
        bus = type(self)()

        for command_cls, handler in self.command_handlers.items():  # type:ignore
            bus.subscribe_command(command_cls, handler)

        for event_cls, handler in self.event_handlers.items():  # type:ignore
            for event_handler in handler:
                bus.subscribe_event(event_cls, event_handler)
        return bus

    def _clear_handlers(self):
        """
        Clear handlers. Mainly for testing purposes.
        """
        self._clear_command_handlers()
        self._clear_event_handlers()

    def _clear_command_handlers(self):
        """
        Clear handlers. Mainly for testing purposes.
        """
        self.command_handlers = {}

    def _clear_event_handlers(self):
        """
        Clear handlers. Mainly for testing purposes.
        """
        self.event_handlers = defaultdict(list)



from mb.messages import Message
from mb.messages import MessageMeta


class EventMeta(MessageMeta):
    """
    The Event metaclass
    """

    _events: dict[str, "EventMeta"] = {}

    def __new__(cls, name, bases, dic):
        # super checks NAME correctness and may assign a default
        event_cls = super().__new__(cls, name, bases, dic)
        cls._events[event_cls.NAME] = event_cls
        return event_cls

    @classmethod
    def _clear(cls):
        """
        Resets internal state. For testing.
        """
        cls._events = {}


class Event(Message, metaclass=EventMeta):
    """
    The Event base class. To inherit and set the NAME.
    """



from mb.messages import Message
from mb.messages import MessageMeta


class CommandMeta(MessageMeta):
    """
    The Command metaclass
    """

    _commands: dict[str, "CommandMeta"] = {}

    def __new__(cls, name, bases, dic):
        # super checks NAME correctness and may assign a default
        command_cls = super().__new__(cls, name, bases, dic)
        cls._commands[command_cls.NAME] = command_cls
        return command_cls

    @classmethod
    def _clear(cls):
        """
        Resets internal state. For testing.
        """
        cls._commands = {}


class Command(Message, metaclass=CommandMeta):
    """
    The Command base class. To inherit and set the NAME.
    """

    # TODO: Could we define a CommandResult inside a Command?
    # In this way, we could use typing to define the output of
    # UnitOfWork.handle_command and preventively detect errors.



import typing as t

from mb.exceptions import CQProgrammingError
from mb.exceptions import UowContextRequired
from mb.unit_of_work.base import UnitOfWork
from mb.unit_of_work.utils.events_collector import DedupeEventsFifo

if t.TYPE_CHECKING:
    from mb.bus import MessageBus
    from mb.events import Event
    from mb.unit_of_work.utils.events_collector import EventsCollector


class NestedUnitOfWork(UnitOfWork):
    """
    Handles events on outermost UOW commit.
    """

    stack: list["Transaction"]
    events_collector_cls: type["EventsCollector"]

    def __init__(
        self,
        bus: "MessageBus",
        events_collector_cls: type["EventsCollector"] = DedupeEventsFifo,
    ):
        super().__init__(bus)
        self.stack = []
        self.events_collector_cls = events_collector_cls

    # Transaction methods

    def __enter__(self):
        self._begin()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not exc_type:
            self._commit()
        else:
            self._rollback()

    def _begin(self):
        transaction = Transaction(
            self,
            self.events_collector_cls(),
            self.stack[-1] if self.stack else None,
        )
        self.stack.append(transaction)

    def _commit(self):
        self._end().commit()

    def _rollback(self):
        self._end().rollback()

    def _end(self) -> "Transaction":
        self._ensure_context()
        return self.stack.pop()

    def _ensure_context(self):
        if not self.stack:
            raise UowContextRequired("No transaction in progress")

    # Events methods

    def _emit_event(self, event: "Event"):
        """
        Collect on the current transaction
        """
        self._ensure_context()
        self.stack[-1].collect_event(event)

    def _handle_events(self, events: t.Iterable["Event"]):
        """
        Called on commit the outermost transaction
        """
        if self.stack:
            raise CQProgrammingError("This call should happen outside the context")
        for event in events:
            self._handle_event(event)


class Transaction:
    """
    The transaction holds the events emmitted during its lifespan and
    eventually calls the handlers on commit. If the transaction is
    rolled back, their events are also discarded.

    Nested transactions do the same but differ on commit. When committing
    a nested transaction, it just passes its events to the parent transaction
    instead of calling the handlers.
    """

    uow: "NestedUnitOfWork"
    events: "EventsCollector"
    parent: t.Optional["Transaction"]

    def __init__(
        self,
        uow: "NestedUnitOfWork",
        events: "EventsCollector",
        parent: "Transaction" = None,
    ):
        self.uow = uow
        self.events = events
        self.parent = parent

    def commit(self):
        if self.parent:
            self.parent.collect_event_many(self.events)
        else:
            self.uow._handle_events(self.events)

    def rollback(self):
        self.events.clear()

    def collect_event(self, event: "Event"):
        self.events.push(event)

    def collect_event_many(self, events: "EventsCollector"):
        self.events.extend(events)
